@@ -1,5 +1,3 @@
-# -*- coding: utf-8 -*-
-#
 # This file is part of the python-chess library.
 # Copyright (C) 2012-2020 Niklas Fiekas <niklas.fiekas@backscattering.de>
 #
@@ -16,6 +14,7 @@
 # You should have received a copy of the GNU General Public License
 # along with this program. If not, see <http://www.gnu.org/licenses/>.
 
+import abc
 import enum
 import itertools
 import logging
@@ -83,7 +82,7 @@ NAG_BLACK_SEVERE_TIME_PRESSURE = 139
 NAG_NOVELTY = 146
 
 
-TAG_REGEX = re.compile(r"^\[([A-Za-z0-9_]+)\s+\"(.*)\"\]\s*$")
+TAG_REGEX = re.compile(r"^\[([A-Za-z0-9_]+)\s+\"([^\r]*)\"\]\s*$")
 
 TAG_NAME_REGEX = re.compile(r"^[A-Za-z0-9_]+\Z")
 
@@ -119,6 +118,16 @@ SKIP = SkipType.SKIP
 ResultT = TypeVar("ResultT", covariant=True)
 
 
+class _AcceptFrame:
+    def __init__(self, node: "GameNode", *, is_variation: bool = False, sidelines: bool = True):
+        assert node.parent is not None, "cannot visit dangling GameNode"
+        self.state = "pre"
+        self.node = node
+        self.is_variation = is_variation
+        self.variations = iter(itertools.islice(node.parent.variations, 1, None) if sidelines else [])
+        self.in_variation = False
+
+
 class GameNode:
     def __init__(self) -> None:
         self.parent: Optional[GameNode] = None
@@ -135,27 +144,44 @@ class GameNode:
     def dangling_node(cls) -> "GameNode":
         return GameNode()
 
-    def board(self, *, _cache: bool = True) -> chess.Board:
+    def _board(self) -> chess.Board:
+        assert False, "cannot get board of dangling GameNode"
+
+    def board(self) -> chess.Board:
         """
         Gets a board with the position of the node.
 
+        For the root node, this is the default starting position (for the
+        ``Variant``) unless the ``FEN`` header tag is set.
+
         It's a copy, so modifying the board will not alter the game.
         """
-        assert self.parent is not None and self.move is not None, "cannot get board of dangling GameNode"
-
         if self.board_cached is not None:
-            board = self.board_cached()
-            if board is not None:
-                return board.copy()
+            board_cached = self.board_cached()
+            if board_cached is not None:
+                return board_cached.copy()
 
-        board = self.parent.board(_cache=False)
-        board.push(self.move)
+        stack = []
+        node = self
+        while True:
+            if node.parent is None or node.move is None:
+                board = node._board()
+                break
+            elif node.board_cached is not None:
+                board_cached = node.board_cached()
+                if board_cached is not None:
+                    board = board_cached.copy()
+                    break
+            stack.append(node.move)
+            node = node.parent
 
-        if _cache:
+        while stack:
+            board.push(stack.pop())
+
+        if self.parent is not None:
             self.board_cached = weakref.ref(board)
-            return board.copy()
-        else:
-            return board
+
+        return board
 
     def _move(self) -> chess.Move:
         assert self.move is not None, "cannot get move of dangling GameNode"
@@ -359,23 +385,38 @@ class GameNode:
             visitor.visit_comment(self.comment)
 
     def _accept(self, parent_board: chess.Board, visitor: "BaseVisitor[ResultT]", *, sidelines: bool = True) -> None:
-        assert self.parent is not None and self.move is not None, "cannot visit dangling GameNode"
+        stack = [_AcceptFrame(self, sidelines=sidelines)]
 
-        # First, visit the move that leads to this node.
-        self._accept_node(parent_board, visitor)
+        while stack:
+            top = stack[-1]
 
-        # Then visit sidelines.
-        if sidelines and self == self.parent.variations[0]:
-            for variation in itertools.islice(self.parent.variations, 1, None):
-                if visitor.begin_variation() is not SKIP:
-                    variation._accept(parent_board, visitor)
+            if top.in_variation:
+                top.in_variation = False
                 visitor.end_variation()
 
-        # The mainline is continued last.
-        if self.variations:
-            parent_board.push(self.move)
-            self.variations[0]._accept(parent_board, visitor)
-            parent_board.pop()
+            if top.state == "pre":
+                top.node._accept_node(parent_board, visitor)
+                top.state = "variations"
+            elif top.state == "variations":
+                try:
+                    variation = next(top.variations)
+                except StopIteration:
+                    if top.node.variations:
+                        assert top.node.move is not None, "root node cannot be a variation"
+                        parent_board.push(top.node.move)
+                        stack.append(_AcceptFrame(top.node.variations[0], sidelines=True))
+                        top.state = "post"
+                    else:
+                        top.state = "end"
+                else:
+                    if visitor.begin_variation() is not SKIP:
+                        stack.append(_AcceptFrame(variation, sidelines=False, is_variation=True))
+                    top.in_variation = True
+            elif top.state == "post":
+                parent_board.pop()
+                top.state = "end"
+            else:
+                stack.pop()
 
     def accept(self, visitor: "BaseVisitor[ResultT]") -> ResultT:
         """
@@ -422,9 +463,13 @@ class GameNode:
 
     def __repr__(self) -> str:
         if self.parent is None:
-            return f"<{type(self).__name__} at {id(self):#x} (dangling)>"
-        else:
+            return f"<{type(self).__name__} at {id(self):#x} (dangling: {self.move})>"
+
+        try:
             parent_board = self.parent.board()
+        except ValueError:
+            return f"<{type(self).__name__} at {id(self):#x} (dangling: {self.move})>"
+        else:
             return "<{} at {:#x} ({}{} {} ...)>".format(
                 type(self).__name__,
                 id(self),
@@ -447,13 +492,7 @@ class Game(GameNode):
         self.headers = Headers(headers)
         self.errors: List[Exception] = []
 
-    def board(self, *, _cache: bool = False) -> chess.Board:
-        """
-        Gets the starting position of the game.
-
-        Unless the ``FEN`` header tag is set, this is the default starting
-        position (for the ``Variant``).
-        """
+    def _board(self) -> chess.Board:
         return self.headers.board()
 
     def setup(self, board: Union[chess.Board, str]) -> None:
@@ -524,7 +563,7 @@ class Game(GameNode):
 
     @classmethod
     def without_tag_roster(cls: Type[GameT]) -> GameT:
-        """Creates an empty game without the default 7 tag roster."""
+        """Creates an empty game without the default Seven Tag Roster."""
         return cls(headers={})
 
     @classmethod
@@ -703,7 +742,7 @@ class ReverseMainline(Generic[MainlineMapT]):
             " ".join(ReverseMainline(self.stop, lambda node: node._move().uci())))
 
 
-class BaseVisitor(Generic[ResultT]):
+class BaseVisitor(abc.ABC, Generic[ResultT]):
     """
     Base class for visitors.
 
@@ -791,9 +830,9 @@ class BaseVisitor(Generic[ResultT]):
         """Called at the end of a game."""
         pass
 
+    @abc.abstractmethod
     def result(self) -> ResultT:
-        """Called to get the result of the visitor. Defaults to ``True``."""
-        return True
+        """Called to get the result of the visitor."""
 
     def handle_error(self, error: Exception) -> None:
         """Called for encountered errors. Defaults to raising an exception."""
@@ -864,7 +903,7 @@ class GameBuilder(BaseVisitor[GameT]):
         Populates :data:`chess.pgn.Game.errors` with encountered errors and
         logs them.
 
-        You can silence the log and handle errors yourself, after parsing:
+        You can silence the log and handle errors yourself after parsing:
 
         >>> import chess.pgn
         >>> import logging
@@ -959,6 +998,8 @@ class SkipVisitor(BaseVisitor["typing.Literal[True]"]):
     def begin_variation(self) -> SkipType:
         return SKIP
 
+    def result(self) -> "typing.Literal[True]":
+        return True
 
 
 class StringExporterMixin:
@@ -1161,7 +1202,7 @@ def read_game(handle: TextIO, *, Visitor = GameBuilder):
     The end of a game is determined by a completely blank line or the end of
     the file. (Of course, blank lines in comments are possible).
 
-    According to the PGN standard, at least the usual 7 header tags are
+    According to the PGN standard, at least the usual seven header tags are
     required for a valid game. This parser also handles games without any
     headers just fine.
 
@@ -1391,7 +1432,7 @@ def read_headers(handle: TextIO) -> Optional[Headers]:
     ...     if "Kasparov" in headers.get("White", "?"):
     ...         kasparov_offsets.append(offset)
 
-    Then it can later be seeked an parsed.
+    Then it can later be seeked and parsed.
 
     >>> for offset in kasparov_offsets:
     ...     pgn.seek(offset)
@@ -1408,6 +1449,6 @@ def read_headers(handle: TextIO) -> Optional[Headers]:
 
 def skip_game(handle: TextIO) -> bool:
     """
-    Skip a game. Returns ``True`` if a game was found and skipped.
+    Skips a game. Returns ``True`` if a game was found and skipped.
     """
     return bool(read_game(handle, Visitor=SkipVisitor))
